import io
import os
import tempfile

import boto3
import numpy as np
from bioio.writers import OmeTiffWriter
from PIL import Image

from io_collection.save.save_buffer import _save_buffer_to_s3

EXTENSIONS = (".ome.tif", ".ome.tiff", ".png")


def save_image(location: str, key: str, image: np.ndarray) -> None:
    """
    Save image array to key at specified location.

    Method will save to the S3 bucket if the location begins with the **s3://**
    protocol, otherwise it assumes the location is a local path.

    Parameters
    ----------
    location
        Object location (local path or S3 bucket).
    key
        Object key ending in `.ome.tiff` or `.png`.
    image
        Image array.
    """

    if not key.endswith(EXTENSIONS):
        raise ValueError(
            f"key [ {key} ] must have [ {' | '.join([ext[1:] for ext in EXTENSIONS])} ] extension"
        )

    if location[:5] == "s3://":
        _save_image_to_s3(location[5:], key, image)
    else:
        _save_image_to_fs(location, key, image)


def _save_image_to_fs(path: str, key: str, image: np.ndarray) -> None:
    """
    Save image array to key on local file system.

    Parameters
    ----------
    path
        Local object path.
    key
        Object key ending in `.ome.tiff` or `.png`.
    image
        Image array.
    """

    full_path = os.path.join(path, key)
    os.makedirs(os.path.split(full_path)[0], exist_ok=True)

    if key.endswith(".ome.tiff") or key.endswith(".ome.tif"):
        OmeTiffWriter.save(image, full_path)
    elif key.endswith(".png"):
        Image.fromarray(image).save(full_path)  # type: ignore


def _save_image_to_s3(bucket: str, key: str, image: np.ndarray) -> None:
    """
    Save image array to key in AWS S3 bucket.

    Parameters
    ----------
    bucket
        AWS S3 bucket name.
    key
        Object key ending in `.ome.tiff` or `.png`.
    image
        Image array.
    """

    s3_client = boto3.client("s3")

    if key.endswith(".ome.tiff") or key.endswith(".ome.tif"):
        with tempfile.TemporaryDirectory() as temp_dir:
            temp_path = os.path.join(temp_dir, "temp.ome.tiff")
            OmeTiffWriter.save(image, temp_path)

            with open(temp_path, "rb") as fileobj:
                s3_client.upload_fileobj(fileobj, bucket, key)
    elif key.endswith(".png"):
        with io.BytesIO() as buffer:
            Image.fromarray(image).save(buffer, format="png")  # type: ignore
            _save_buffer_to_s3(bucket, key, buffer, "image/png")



"""Tasks for working with object keys."""

from prefect import task

from .change_key import change_key
from .check_key import check_key
from .copy_key import copy_key
from .get_keys import get_keys
from .make_key import make_key
from .remove_key import remove_key

change_key = task(change_key)
check_key = task(check_key)
copy_key = task(copy_key)
get_keys = task(get_keys)
make_key = task(make_key)
remove_key = task(remove_key)



import io
import pickle
from typing import Any

from io_collection.save.save_buffer import save_buffer


def save_pickle(location: str, key: str, obj: Any) -> None:
    """
    Save pickled object to key at specified location.

    Method will save to the S3 bucket if the location begins with the **s3://**
    protocol, otherwise it assumes the location is a local path.

    Parameters
    ----------
    location
        Object location (local path or S3 bucket).
    key
        Object key ending in `.pkl`.
    obj
        Object to save.
    """

    if not key.endswith(".pkl"):
        raise ValueError(f"key [ {key} ] must have [ pkl ] extension")

    with io.BytesIO() as buffer:
        pickle.dump(obj, buffer)
        save_buffer(location, key, buffer)



import io
import os
from typing import Any

import pandas as pd

from io_collection.save.save_buffer import _save_buffer_to_s3


def save_dataframe(location: str, key: str, dataframe: pd.DataFrame, **kwargs: Any) -> None:
    """
    Save dataframe to key at specified location.

    Method will save to the S3 bucket if the location begins with the **s3://**
    protocol, otherwise it assumes the location is a local path.

    Parameters
    ----------
    location
        Object location (local path or S3 bucket).
    key
        Object key ending in `.csv`.
    dataframe
        Dataframe to save.
    **kwargs
        Additional parameters for saving dataframe. The keyword arguments are
        passed to `pandas.to_csv`.
    """

    if not key.endswith(".csv"):
        raise ValueError(f"key [ {key} ] must have [ csv ] extension")

    if location[:5] == "s3://":
        _save_dataframe_to_s3(location[5:], key, dataframe, **kwargs)
    else:
        _save_dataframe_to_fs(location, key, dataframe, **kwargs)


def _save_dataframe_to_fs(path: str, key: str, dataframe: pd.DataFrame, **kwargs: Any) -> None:
    """
    Save dataframe to key on local file system.

    Parameters
    ----------
    path
        Local object path.
    key
        Object key ending in `.csv`.
    dataframe
        Dataframe to save.
    **kwargs
        Additional parameters for saving dataframe. The keyword arguments are
        passed to `pandas.to_csv`.
    """

    full_path = os.path.join(path, key)
    os.makedirs(os.path.split(full_path)[0], exist_ok=True)
    dataframe.to_csv(full_path, **kwargs)


def _save_dataframe_to_s3(bucket: str, key: str, dataframe: pd.DataFrame, **kwargs: Any) -> None:
    """
    Save dataframe to key in AWS S3 bucket.

    Parameters
    ----------
    bucket
        AWS S3 bucket name.
    key
        Object key ending in `.csv`.
    dataframe
        Dataframe to save.
    **kwargs
        Additional parameters for saving dataframe. The keyword arguments are
        passed to `pandas.to_csv`.
    """

    with io.BytesIO() as buffer:
        dataframe.to_csv(buffer, **kwargs)
        _save_buffer_to_s3(bucket, key, buffer, "text/csv")



import io
import os

import boto3


def save_buffer(
    location: str, key: str, buffer: io.BytesIO, content_type: str = "binary/octet-stream"
) -> None:
    """
    Save buffer to key at specified location.

    Method will save to the S3 bucket if the location begins with the
    **s3://** protocol, otherwise it assumes the location is a local path.

    Parameters
    ----------
    location
        Object location (local path or S3 bucket).
    key
        Object key.
    buffer
        Content buffer.
    content_type
        Content type (S3 only).
    """

    if location[:5] == "s3://":
        _save_buffer_to_s3(location[5:], key, buffer, content_type)
    else:
        _save_buffer_to_fs(location, key, buffer)


def _save_buffer_to_fs(path: str, key: str, buffer: io.BytesIO) -> None:
    """
    Save buffer to key on local file system.

    Directories that do not exist will be created.

    Parameters
    ----------
    path
        Local object path.
    key
        Object key.
    buffer
        Content buffer.
    """

    full_path = os.path.join(path, key)
    os.makedirs(os.path.split(full_path)[0], exist_ok=True)
    with open(full_path, "wb") as file:
        file.write(buffer.getvalue())


def _save_buffer_to_s3(bucket: str, key: str, buffer: io.BytesIO, content_type: str) -> None:
    """
    Save buffer to key in AWS S3 bucket.

    Parameters
    ----------
    bucket
        AWS S3 bucket name.
    key
        Object key.
    buffer
        Content buffer.
    content_type
        Content type.
    """

    s3_client = boto3.client("s3")
    s3_client.put_object(Bucket=bucket, Key=key, Body=buffer.getvalue(), ContentType=content_type)



import io
import os

import boto3

MAX_CONTENT_LENGTH = 2**31 - 1


def load_buffer(location: str, key: str) -> io.BytesIO:
    """
    Load key into in-memory bytes buffer from specified location.

    Method will load from the S3 bucket if the location begins with the
    **s3://** protocol, otherwise it assumes the location is a local path.

    Parameters
    ----------
    location
        Object location (local path or S3 bucket).
    key
        Object key.

    Returns
    -------
    :
        Loaded object buffer.
    """

    if location[:5] == "s3://":
        return _load_buffer_from_s3(location[5:], key)
    return _load_buffer_from_fs(location, key)


def _load_buffer_from_fs(path: str, key: str) -> io.BytesIO:
    """
    Load key into in-memory bytes buffer from local file system.

    Parameters
    ----------
    path
        Local object path.
    key
        Object key.

    Returns
    -------
    :
        Loaded object buffer.
    """

    full_path = os.path.join(path, key)
    return io.BytesIO(open(full_path, "rb").read())


def _load_buffer_from_s3(bucket: str, key: str) -> io.BytesIO:
    """
    Load key into in-memory bytes buffer from AWS S3 bucket.

    Objects larger than `MAX_CONTENT_LENGTH` are loaded in chunks.

    Parameters
    ----------
    bucket
        AWS S3 bucket name.
    key
        Object key.

    Returns
    -------
    :
        Loaded object buffer.
    """

    s3_client = boto3.client("s3")
    obj = s3_client.get_object(Bucket=bucket, Key=key)

    # Check if body needs to be loaded in chunks.
    content_length = obj["ContentLength"]

    if content_length > MAX_CONTENT_LENGTH:
        print("Loading chunks ...")
        body = bytearray()
        for chunk in obj["Body"].iter_chunks(chunk_size=MAX_CONTENT_LENGTH):
            body += chunk
        return io.BytesIO(body)

    return io.BytesIO(obj["Body"].read())



import os
from datetime import datetime


def make_key(*subkeys: str) -> str:
    """
    Combines given subkeys into a single key.

    If any subkeys include **{{timestamp}}**, it will be replaced with the
    current date formatted as **YYYY-MM-DD**. Any instances of double
    underscores (`__`) are replaced with a single underscore (`_`). Any
    instances of underscore followed by a period (`_.`) are replaced with a
    period (`.`).

    Returns
    -------
    :
        The key.
    """

    key = os.path.join(*subkeys)

    timestamp = datetime.now().strftime("%Y-%m-%d")
    key = key.replace("{{timestamp}}", timestamp)

    key = key.replace("__", "_")
    key = key.replace("_.", ".")

    return key



import io
import os

from io_collection.save.save_buffer import _save_buffer_to_s3


def save_text(
    location: str, key: str, text: str, content_type: str = "binary/octet-stream"
) -> None:
    """
    Save text to key at specified location.

    Method will save to the S3 bucket if the location begins with the **s3://**
    protocol, otherwise it assumes the location is a local path.

    Parameters
    ----------
    location
        Object location (local path or S3 bucket).
    key
        Object key ending in valid text extension.
    text
        Text to save.
    content_type
        Content type (S3 only).
    """

    if location[:5] == "s3://":
        _save_text_to_s3(location[5:], key, text, content_type)
    else:
        _save_text_to_fs(location, key, text)


def _save_text_to_fs(path: str, key: str, text: str) -> None:
    """
    Save text to key on local file system.

    Parameters
    ----------
    path
        Local object path.
    key
        Object key ending in valid text extension.
    text
        Text to save.
    """

    full_path = os.path.join(path, key)
    os.makedirs(os.path.split(full_path)[0], exist_ok=True)
    with open(full_path, "w", encoding="utf-8") as file:
        file.write(text)


def _save_text_to_s3(bucket: str, key: str, text: str, content_type: str) -> None:
    """
    Save text to key in AWS S3 bucket.

    Parameters
    ----------
    bucket
        AWS S3 bucket name.
    key
        Object key ending in valid text extension.
    text
        Text to save.
    content_type
        Content type (S3 only).
    """

    with io.BytesIO() as buffer:
        buffer.write(text.encode("utf-8"))
        _save_buffer_to_s3(bucket, key, buffer, content_type)



import quilt3


def load_quilt_package(quilt_package: str, quilt_registry: str) -> quilt3.Package:
    """
    Load Quilt package into memory from registry.

    Package is loaded without making a local copy of the manifest.

    Parameters
    ----------
    quilt_package
        Quilt package name, formatted as {namespace}/{packagename}.
    quilt_registry
        Package registry name.

    Returns
    -------
    :
        The Quilt package.
    """

    package = quilt3.Package.browse(quilt_package, quilt_registry)
    return package



import os

import boto3


def remove_key(location: str, key: str) -> None:
    """
    Removes object key at specified location.

    Method will remove from the S3 bucket if the location begins with the
    **s3://** protocol, otherwise it assumes the location is a local path.

    Parameters
    ----------
    location
        Object location (local path or S3 bucket).
    key
        Object key.
    """

    if location[:5] == "s3://":
        _remove_key_on_s3(location[5:], key)
    else:
        _remove_key_on_fs(location, key)


def _remove_key_on_fs(path: str, key: str) -> None:
    """
    Remove object key from local file system.

    Parameters
    ----------
    path
        Local object path.
    key
        Object key.
    """

    full_path = os.path.join(path, key)

    if os.path.isfile(full_path):
        os.remove(full_path)


def _remove_key_on_s3(bucket: str, key: str) -> None:
    """
    Remove object key in AWS S3 bucket.

    Parameters
    ----------
    path
        AWS S3 bucket name.
    key
        Object key.
    """

    s3_client = boto3.client("s3")
    s3_client.delete_object(Bucket=bucket, Key=key)



"""Tasks for loading and saving objects via Quilt."""

from prefect import task

from .load_quilt_package import load_quilt_package
from .save_quilt_item import save_quilt_item

load_quilt_package = task(load_quilt_package)
save_quilt_item = task(save_quilt_item)



import os

from io_collection.load.load_buffer import _load_buffer_from_s3


def load_text(location: str, key: str) -> str:
    """
    Load key as string from specified location.

    Method will load from the S3 bucket if the location begins with the
    **s3://** protocol, otherwise it assumes the location is a local path.

    Parameters
    ----------
    location
        Object location (local path or S3 bucket).
    key
        Object key ending in valid text extension.

    Returns
    -------
    :
        Loaded text.
    """

    if location[:5] == "s3://":
        return _load_text_from_s3(location[5:], key)
    return _load_text_from_fs(location, key)


def _load_text_from_fs(path: str, key: str) -> str:
    """
    Load key as dict or list from local file system.

    Parameters
    ----------
    path
        Local object path.
    key
        Object key ending in valid text extension.

    Returns
    -------
    :
        Loaded text.
    """

    full_path = os.path.join(path, key)
    return open(full_path, "r", encoding="utf-8").read()


def _load_text_from_s3(bucket: str, key: str) -> str:
    """
    Load key as dict or list from AWS S3 bucket.

    Parameters
    ----------
    bucket
        AWS S3 bucket name.
    key
        Object key ending in valid text extension.

    Returns
    -------
    :
        Loaded text.
    """

    buffer = _load_buffer_from_s3(bucket, key)
    return buffer.getvalue().decode("utf-8")



import io
import os
from typing import Any

import matplotlib.figure as mpl

from io_collection.save.save_buffer import _save_buffer_to_s3

EXTENSIONS = (".png", ".jpeg", ".jpg", ".svg")

CONTENT_TYPES = {
    "png": "image/png",
    "jpeg": "image/jpeg",
    "jpg": "image/jpeg",
    "svg": "image/svg+xml",
}


def save_figure(location: str, key: str, figure: mpl.Figure, **kwargs: Any) -> None:
    """
    Save matplotlib figure to key at specified location.

    Method will save to the S3 bucket if the location begins with the **s3://**
    protocol, otherwise it assumes the location is a local path.

    Parameters
    ----------
    location
        Object location (local path or S3 bucket).
    key
        Object key ending in `.png`, `.jpg`, or `.svg`.
    figure
        Figure instance to save.
    **kwargs
        Additional parameters for saving figure. The keyword arguments are
        passed to `matplotlib.pyplot.savefig`.
    """

    if not key.endswith(EXTENSIONS):
        raise ValueError(
            f"key [ {key} ] must have [ {' | '.join([ext[1:] for ext in EXTENSIONS])} ] extension"
        )

    if location[:5] == "s3://":
        _save_figure_to_s3(location[5:], key, figure, **kwargs)
    else:
        _save_figure_to_fs(location, key, figure, **kwargs)


def _save_figure_to_fs(path: str, key: str, figure: mpl.Figure, **kwargs: Any) -> None:
    """
    Save matplotlib figure to key on local file system.

    Parameters
    ----------
    path
        Local object path.
    key
        Object key ending in `.png`, `.jpg`, or `.svg`.
    figure
        Figure instance to save.
    **kwargs
        Additional parameters for saving figure. The keyword arguments are
        passed to `matplotlib.pyplot.savefig`.
    """

    full_path = os.path.join(path, key)
    os.makedirs(os.path.split(full_path)[0], exist_ok=True)
    figure.savefig(full_path, **kwargs)


def _save_figure_to_s3(bucket: str, key: str, figure: mpl.Figure, **kwargs: Any) -> None:
    """
    Save matplotlib figure to key in AWS S3 bucket.

    Parameters
    ----------
    bucket
        AWS S3 bucket name.
    key
        Object key ending in `.png`, `.jpg`, or `.svg`.
    figure
        Figure instance to save.
    **kwargs
        Additional parameters for saving figure. The keyword arguments are
        passed to `matplotlib.pyplot.savefig`.
    """

    with io.BytesIO() as buffer:
        figure.savefig(buffer, **kwargs)
        content_type = CONTENT_TYPES[key.split(".")[-1]]
        _save_buffer_to_s3(bucket, key, buffer, content_type)



import pickle
from typing import Any

from io_collection.load.load_buffer import load_buffer


def load_pickle(location: str, key: str) -> Any:
    """
    Load key as pickled object from specified location.

    Method will load from the S3 bucket if the location begins with the
    **s3://** protocol, otherwise it assumes the location is a local path.

    Parameters
    ----------
    location
        Object location (local path or S3 bucket).
    key
        Object key ending in `.pkl`.

    Returns
    -------
    :
        Loaded object.
    """

    if not key.endswith(".pkl"):
        raise ValueError(f"key [ {key} ] must have [ pkl ] extension")

    buffer = load_buffer(location, key)
    return pickle.loads(buffer.read())



import os

import boto3


def change_key(location: str, old_key: str, new_key: str) -> None:
    """
    Change object key at specified location.

    Method will change the object in an S3 bucket if the location begins with
    the **s3://** protocol, otherwise it assumes the location is a local path.

    Parameters
    ----------
    location
        Object location (local path or S3 bucket).
    old_key
        Old object key.
    new_key
        New object key.
    """

    if location[:5] == "s3://":
        _change_key_on_s3(location[5:], old_key, new_key)
    else:
        _change_key_on_fs(location, old_key, new_key)


def _change_key_on_fs(path: str, old_key: str, new_key: str) -> None:
    """
    Change object key on local file system.

    Parameters
    ----------
    path
        Local object path.
    old_key
        Old object key.
    new_key
        New object key.
    """

    full_old_path = os.path.join(path, old_key)
    full_new_path = os.path.join(path, new_key)
    os.renames(full_old_path, full_new_path)


def _change_key_on_s3(bucket: str, old_key: str, new_key: str) -> None:
    """
    Change object key in AWS S3 bucket.

    Parameters
    ----------
    path
        AWS S3 bucket name.
    old_key
        Old object key.
    new_key
        New object key.
    """

    s3_client = boto3.client("s3")
    s3_client.copy_object(Bucket=bucket, CopySource=f"{bucket}/{old_key}", Key=new_key)
    s3_client.delete_object(Bucket=bucket, Key=old_key)



import io

import quilt3

from io_collection.save.save_buffer import save_buffer


def save_quilt_item(location: str, key: str, package: quilt3.Package, item: str) -> None:
    """
    Save item from Quilt package to key at specified location.

    Parameters
    ----------
    location
        Object location (local path or S3 bucket).
    key
        Object key.
    package
        Quilt package.
    item
        Name of the item in the package.
    """

    contents = package[item].get_bytes()
    buffer = io.BytesIO(contents)
    save_buffer(location, key, buffer)



"""Tasks for saving different object types to local file systems or S3 buckets."""

from prefect import task

from .save_buffer import save_buffer
from .save_dataframe import save_dataframe
from .save_figure import save_figure
from .save_gif import save_gif
from .save_json import save_json
from .save_pickle import save_pickle
from .save_tar import save_tar
from .save_text import save_text

save_buffer = task(save_buffer)
save_dataframe = task(save_dataframe)
save_figure = task(save_figure)
save_gif = task(save_gif)
save_json = task(save_json)
save_pickle = task(save_pickle)
save_tar = task(save_tar)
save_text = task(save_text)

try:
    from .save_image import save_image

    save_image = task(save_image)
except ImportError:
    pass



import io

from PIL import Image

from io_collection.load.load_buffer import load_buffer
from io_collection.save.save_buffer import save_buffer


def save_gif(location: str, key: str, frame_keys: list[str]) -> None:
    """
    Save series of images as gif to key at specified location.

    Method will save to the S3 bucket if the location begins with the **s3://**
    protocol, otherwise it assumes the location is a local path.

    Parameters
    ----------
    location
        Object location (local path or S3 bucket).
    key
        Object key ending in `.gif`.
    contents
        List of frame keys to include in the gif.
    """

    if not key.endswith(".gif"):
        raise ValueError(f"key [ {key} ] must have [ gif ] extension")

    with io.BytesIO() as buffer:
        frames = [Image.open(load_buffer(location, frame_key)) for frame_key in frame_keys]
        frames[0].save(
            buffer, format="gif", save_all=True, append_images=frames[1:], duration=100, loop=0
        )
        save_buffer(location, key, buffer, "image/gif")



"""Tasks for loading different object types from local file systems or S3 buckets."""

from prefect import task

from .load_buffer import load_buffer
from .load_dataframe import load_dataframe
from .load_json import load_json
from .load_pickle import load_pickle
from .load_tar import load_tar
from .load_text import load_text

load_buffer = task(load_buffer)
load_dataframe = task(load_dataframe)
load_json = task(load_json)
load_pickle = task(load_pickle)
load_tar = task(load_tar)
load_text = task(load_text)

try:
    from .load_image import load_image

    load_image = task(load_image)
except ImportError:
    pass



import json
import re
from typing import Union

from io_collection.save.save_text import save_text


def save_json(location: str, key: str, obj: Union[dict, list], levels: int = 5) -> None:
    """
    Save dict or list as json to key at specified location.

    Method will save to the S3 bucket if the location begins with the **s3://**
    protocol, otherwise it assumes the location is a local path.

    The json is pretty printed with an indentation level of 2. For lists,
    default pretty printing places each item on a new line which can produce
    very long files. This method will attempt map lists up to the specified
    number of items onto a single line.

    Parameters
    ----------
    location
        Object location (local path or S3 bucket).
    key
        Object key ending in `.json`.
    obj
        Dict or list to save.
    levels
        Maximum number of items for single line list.
    """

    if not key.endswith(".json"):
        raise ValueError(f"key [ {key} ] must have [ json ] extension")

    contents = json.dumps(obj, indent=2)

    for level in range(1, levels + 1):
        match_pattern = (
            r"\[" + ",".join([r"\n\s*([A-z\"\-\d\.]+)" for i in range(level)]) + r"\n\s*\]"
        )
        sub_pattern = "[" + ",".join([f"\\{i + 1}" for i in range(level)]) + "]"
        pattern = re.compile(match_pattern)
        contents = pattern.sub(sub_pattern, contents)

    save_text(location, key, contents, "application/json")



import os

import boto3


def check_key(location: str, key: str) -> bool:
    """
    Check if object key exists at specified location.

    Method will check in the S3 bucket if the location begins with the **s3://**
    protocol, otherwise it assumes the location is a local path.

    Parameters
    ----------
    location
        Object location (local path or S3 bucket).
    key
        Object key.

    Returns
    -------
    :
        True if the object exists in the location, False otherwise.
    """

    if location[:5] == "s3://":
        return _check_key_on_s3(location[5:], key)
    return _check_key_on_fs(location, key)


def _check_key_on_fs(path: str, key: str) -> bool:
    """
    Check if object key exists on local file system.

    Parameters
    ----------
    path
        Local object path.
    key
        Object key.

    Returns
    -------
    bool
        True if the object exists on the local file system, False otherwise.
    """

    full_path = os.path.join(path, key)
    return os.path.isfile(full_path)


def _check_key_on_s3(bucket: str, key: str) -> bool:
    """
    Check if object key exists in AWS S3 bucket.

    Parameters
    ----------
    path
        AWS S3 bucket name.
    key
        Object key.

    Returns
    -------
    bool
        True if the object exists in the AWS bucket, False otherwise.
    """

    s3_client = boto3.client("s3")
    result = s3_client.list_objects(Bucket=bucket, Prefix=key)
    return "Contents" in result



import io
import tarfile

from io_collection.keys.check_key import check_key
from io_collection.load.load_buffer import load_buffer
from io_collection.load.load_tar import load_tar
from io_collection.save.save_buffer import save_buffer


def save_tar(location: str, key: str, contents: list[str]) -> None:
    """
    Save tar archive to key at specified location.

    Method will save to the S3 bucket if the location begins with the **s3://**
    protocol, otherwise it assumes the location is a local path.

    If the archive already exists, all objects in the existing archive will be
    copied into the new archive, along with the new contents.

    Method currently only supports `xz` compression.

    Parameters
    ----------
    location
        Object location (local path or S3 bucket).
    key
        Object key ending in `.tar.xz`.
    contents
        List of content keys to include in the archive.
    """

    if not key.endswith(".tar.xz"):
        raise ValueError(f"key [ {key} ] must have [ tar.xz ] extension")

    if check_key(location, key):
        existing_tar = load_tar(location, key)
    else:
        existing_tar = None

    with io.BytesIO() as buffer:
        with tarfile.open(fileobj=buffer, mode="w:xz") as tar:
            if existing_tar is not None:
                for member in existing_tar.getmembers():
                    tar.addfile(member, existing_tar.extractfile(member.name))

            for content_key in contents:
                content = load_buffer(location, content_key)
                info = tarfile.TarInfo(content_key.split("/")[-1])
                info.size = content.getbuffer().nbytes
                tar.addfile(info, fileobj=content)

        save_buffer(location, key, buffer)



import os
from typing import Optional

import bioio_ome_tiff
import bioio_tifffile
from bioio import BioImage

EXTENSIONS = (".tif", ".tiff", ".ome.tif", ".ome.tiff")


def load_image(location: str, key: str, dim_order: Optional[str] = None) -> BioImage:
    """
    Load key as BioIO image from specified location.

    Method will load from the S3 bucket if the location begins with the
    **s3://** protocol, otherwise it assumes the location is a local path.

    Method currently only supports `.tiff` and `ome.tiff` images.

    Parameters
    ----------
    location
        Object location (local path or S3 bucket).
    key
        Object key ending in `.tiff` or `ome.tiff`.

    Returns
    -------
    :
        Loaded image.
    """

    if not key.endswith(EXTENSIONS):
        raise ValueError(
            f"key [ {key} ] must have [ {' | '.join([ext[1:] for ext in EXTENSIONS])} ] extension"
        )

    if location[:5] == "s3://":
        return _load_image_from_s3(location[5:], key, dim_order)
    return _load_image_from_fs(location, key, dim_order)


def _load_image_from_fs(path: str, key: str, dim_order: Optional[str] = None) -> BioImage:
    """
    Load key as BioIO image from local file system.

    Parameters
    ----------
    path
        Local object path.
    key
        Object key ending in `.tiff` or `ome.tiff`.

    Returns
    -------
    :
        Loaded image.
    """

    full_path = os.path.join(path, key)
    return BioImage(
        full_path,
        reader=bioio_ome_tiff.Reader if ".ome.tif" in key else bioio_tifffile.Reader,
        dim_order=dim_order,
    )


def _load_image_from_s3(bucket: str, key: str, dim_order: Optional[str] = None) -> BioImage:
    """
    Load key as BioIO image from AWS S3 bucket.

    Parameters
    ----------
    bucket
        AWS S3 bucket name.
    key
        Object key ending in `.tiff` or `ome.tiff`.

    Returns
    -------
    :
        Loaded image.
    """

    full_key = f"s3://{bucket}/{key}"
    return BioImage(
        full_key,
        reader=bioio_ome_tiff.Reader if ".ome.tif" in key else bioio_tifffile.Reader,
        dim_order=dim_order,
    )



import os
from typing import Any

import pandas as pd

from io_collection.load.load_buffer import _load_buffer_from_s3


def load_dataframe(location: str, key: str, **kwargs: Any) -> pd.DataFrame:
    """
    Load key as dataframe from specified location.

    Method will load from the S3 bucket if the location begins with the
    **s3://** protocol, otherwise it assumes the location is a local path.

    Parameters
    ----------
    location
        Object location (local path or S3 bucket).
    key
        Object key ending in `.csv`.
    **kwargs
        Additional parameters for loading dataframe. The keyword arguments are
        passed to `pandas.read_csv`.

    Returns
    -------
    :
        Loaded dataframe.
    """

    if not key.endswith(".csv"):
        raise ValueError(f"key [ {key} ] must have [ csv ] extension")

    if location[:5] == "s3://":
        return _load_dataframe_from_s3(location[5:], key, **kwargs)
    return _load_dataframe_from_fs(location, key, **kwargs)


def _load_dataframe_from_fs(path: str, key: str, **kwargs: Any) -> pd.DataFrame:
    """
    Load key as dataframe from local file system.

    Parameters
    ----------
    path
        Local object path.
    key
        Object key ending in `.csv`.
    **kwargs
        Additional parameters for loading dataframe. The keyword arguments are
        passed to `pandas.read_csv`.

    Returns
    -------
    :
        Loaded dataframe.
    """

    full_path = os.path.join(path, key)
    return pd.read_csv(full_path, **kwargs)


def _load_dataframe_from_s3(bucket: str, key: str, **kwargs: Any) -> pd.DataFrame:
    """
    Load key as dataframe from AWS S3 bucket.

    Parameters
    ----------
    bucket
        AWS S3 bucket name.
    key
        Object key ending in `.csv`.
    **kwargs
        Additional parameters for loading dataframe. The keyword arguments are
        passed to `pandas.read_csv`.

    Returns
    -------
    :
        Loaded dataframe.
    """

    buffer = _load_buffer_from_s3(bucket, key)
    return pd.read_csv(buffer, **kwargs)



import os
from glob import glob

import boto3


def get_keys(location: str, prefix: str) -> list[str]:
    """
    Get list of objects at specified location with given prefix.

    Method will check in the S3 bucket if the location begins with the **s3://**
    protocol, otherwise it assumes the location is a local path.

    Parameters
    ----------
    location
        Object location (local path or S3 bucket).
    prefix
        Object key prefix.

    Returns
    -------
    :
        List of all object keys at location.
    """

    if location[:5] == "s3://":
        return _get_keys_on_s3(location[5:], prefix)
    return _get_keys_on_fs(location, prefix)


def _get_keys_on_fs(path: str, prefix: str) -> list[str]:
    """
    Get list of objects on local file system with given prefix.

    Parameters
    ----------
    location
        Local object path.
    prefix
        Object key prefix.

    Returns
    -------
    :
        List of all object keys on the local file system.
    """

    glob_pattern = f"{path}/{prefix}/**/*".replace("//", "/")
    all_files = glob(glob_pattern, recursive=True)
    regular_files = [file for file in all_files if not os.path.isdir(file)]
    keys = [file.replace(path, "").strip("/") for file in regular_files]
    return keys


def _get_keys_on_s3(bucket: str, prefix: str) -> list[str]:
    """
    Get list of objects in AWS S3 bucket with given prefix.

    Parameters
    ----------
    location
        AWS S3 bucket name.
    prefix
        Object key prefix.

    Returns
    -------
    :
        List of all object keys in the AWS bucket.
    """

    s3_client = boto3.client("s3")

    bucket = bucket.replace("s3://", "")
    prefix = f"{prefix}/"
    response = s3_client.list_objects_v2(Bucket=bucket, Prefix=prefix)

    get_response = True
    keys: list[str] = []

    while get_response:
        if "Contents" not in response:
            break

        content_keys = [content["Key"] for content in response["Contents"]]
        keys = keys + content_keys

        if response["IsTruncated"]:
            response = s3_client.list_objects_v2(
                Bucket=bucket,
                Prefix=prefix,
                ContinuationToken=response["NextContinuationToken"],
            )
        else:
            get_response = False

    return keys
